#!/usr/bin/env python
# coding: utf-8

"""
This module contains Jupyter widgets for interactively specifying preferences
as well as interactively and non-interactively displaying solutions.
"""

from .widgets import VegaWidget, NimbusPrefWidget, ParplotWidget

__all__ = ['VegaWidget', 'NimbusPrefWidget', 'ParplotWidget']



#!/usr/bin/python
from matplotlib.ticker import MaxNLocator
from copy import deepcopy


def ticks(vmin, vmax):
    nbins = 9
    steps = [1, 2, 2.5, 5, 10]
    locator = MaxNLocator(
        nbins=nbins,
        steps=steps,
        prune='both')
    return locator.tick_values(vmin, vmax)


def vega3_parplot_df_spec(
        heading_df,
        values_df,
        custom_axis_values=True,
        dim_labels=True,
        dim_tooltips=False,
        dim_symbols=False):
    # Based on https://vega.github.io/vega/examples/parallel-coordinates/
    from altair.vega.data import to_values

    def min_max(row):
        if row['ideal'] < row['nadir']:
            return row['ideal'], row['nadir']
        else:
            return row['nadir'], row['ideal']

    def scale_of_row(row):
        mn, mx = min_max(row)
        return {
            "name": row['name'],
            "type": "linear",
            "range": "height",
            "zero": False,
            "domain": [mn, mx],
        }

    def axis_of_row(row):
        json = {
            "orient": "left",
            "zindex": 1,
            "scale": row['name'],
            "title": row['name'],
            "offset": {
                "scale": "ord",
                "value": row['name'],
                "mult": -1
            }
        }
        if custom_axis_values:
            mn, mx = min_max(row)
            tick_vals = list(ticks(mn, mx))
            json['values'] = (
                [row['ideal']]
                + tick_vals
                + [mx])
            json['format'] = ".5"
        return json

    def text():
        return {
            "signal": "format(parent[datum.data], \".5\")"
        }

    def full_text():
        return {
            "signal": "parent[datum.data]"
        }

    def mark(type):
        json = {
            "type": type,
            "from": {"data": "fields"},
            "encode": {
                "enter": {
                    "x": {"scale": "ord", "field": "data"},
                    "y": {
                        "scale": {"datum": "data"},
                        "field": {"parent": {
                            "datum": "data"
                         }}
                    },
                }
            }
        }
        if type == 'text':
            json['encode']['enter']['text'] = text()
            json['encode']['enter']['x']['offset'] = 4
        if type == 'line':
            json['encode']['enter'].update({
                "stroke": {"value": "red"},
                "strokeWidth": {"value": 1.01},
                "strokeOpacity": {"value": 1}
            })
        if type == 'symbol':
            json['encode']['enter']["stroke"] = {
                "value": "black"
            }
        return json

    def extra_marks():
        extra_marks = []
        if dim_labels:
            extra_marks.append(mark('text'))
        if dim_symbols:
            extra_marks.append(mark('symbol'))
        if dim_tooltips:
            if len(extra_marks) == 0:
                extra_marks.append(mark('path'))
            extra_marks[-1]['encode']['enter']["tooltip"] = full_text()
        return extra_marks

    json = {
        "$schema": "https://vega.github.io/schema/vega/v3.json",
        "width": 400,
        "height": 200,
        "padding": 5,
        "config": {
            "axisY": {
                "titleX": -5,
                "titleY": 205,
                "titleAngle": -30,
                "titleAlign": "right",
                "titleBaseline": "top"
            }
        },
        "data": [
            {
                "name": "main",
                "values": to_values(values_df)['values']
            },
            {
                "name": "fields",
                "values": list(heading_df.name)
            }
        ],
        "scales": [
            {
                "name": "ord",
                "type": "point",
                "range": "width",
                "domain": {
                    "data": "fields",
                    "field": "data"
                }
            }
        ] + [
            scale_of_row(row)
            for idx, row in heading_df.iterrows()
        ],
        "axes": [
            axis_of_row(row)
            for idx, row in heading_df.iterrows()
        ],
        "marks": [
            {
                "type": "group",
                "clip": {"path": "M -40 -20 H 500 V 220 H -40 Z"},
                "from": {"data": "main"},
                "marks": [mark("line")] + extra_marks()
            },
        ]
    }
    return json


def vega3_parplot_spec(results, problem, *args, **kwargs):
    if 'max_as_min' in kwargs:
        max_as_min = kwargs['max_as_min']
        del kwargs['max_as_min']
    else:
        max_as_min = True
    heading_df, values_df = prepare_dfs(results, problem, max_as_min=max_as_min)
    return vega3_parplot_df_spec(heading_df, values_df, *args, **kwargs)


def vega3_parplot(results, problem):
    """
    Plot a parallel cordinate plot of solutions using Altair and Vega 3.

    Parameters
    ----------
    results
        The solutions to plot.

    problem
        The DESDEO problem with which the plot is made with respect to.
    """

    import altair.vega.v3 as vg
    return vg.vega(vega3_parplot_spec(results, problem), validate=True)


def prepare_dfs(results, problem, max_as_min=True):
    import pandas as pd
    results = deepcopy(results)
    ideal = problem.ideal.copy()
    nadir = problem.nadir.copy()

    def adjust(arr):
        for max, (i, v) in zip(problem.maximized, enumerate(arr)):
            if not max:
                continue
            arr[i] = -v

    if not max_as_min:
        for rec in results:
            adjust(rec)
        adjust(ideal)
        adjust(nadir)

    return (
        pd.DataFrame({
            'name': problem.objectives,
            'ideal': ideal,
            'nadir': nadir
        }),
        pd.DataFrame.from_records(
            results,
            columns=problem.objectives,
        ),
    )



"""
This module contains the machinery for doing non-interactive plotting of
multidimensional data. Currently it can generate VEGA specifications for
parallel coordinates plots.
"""
__all__ = ['parplot']

from .parallel import vega3_parplot as parplot



#!/usr/bin/env python
# coding: utf-8

from ipywidgets import DOMWidget
from traitlets import Unicode, Dict, List, Bool, observe, default
from desdeo.utils.exceptions import DESDEOException
from desdeo.method.NIMBUS import NIMBUS
from desdeo.preference import NIMBUSClassification
from desdeo_vis._version import EXTENSION_SPEC_VERSION
from desdeo_vis.conf import get_conf

module_name = "desdeo_vis"


class InvalidNimbusPreferencesException(DESDEOException):
    pass


class VegaWidget(DOMWidget):
    _model_name = Unicode('VegaModel').tag(sync=True)
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(EXTENSION_SPEC_VERSION).tag(sync=True)
    _view_name = Unicode('VegaView').tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(EXTENSION_SPEC_VERSION).tag(sync=True)

    spec = Dict().tag(sync=True)


class ParplotWidget(VegaWidget):
    """
    A parallel plot widget to display solutions to a multi-objective
    optimization problem.
    """

    _model_name = Unicode('ParplotModel').tag(sync=True)
    _view_name = Unicode('ParplotView').tag(sync=True)

    maximized = List().tag(sync=True)
    orig_max_as_min = Bool().tag(sync=True)
    cur_max_as_min = Bool().tag(sync=True)

    def __init__(self, results, problem, max_as_min=None, **kwargs):
        """
        Parameters
        ----------
        results
            The solutions to plot.

        problem
            The DESDEO problem with which the plot is made with respect to.

        max_as_min
            Whether to reformulate maximized functions as minimized functions.
        """
        self.results = results
        self.problem = problem
        self.maximized = problem.maximized
        if max_as_min is not None:
            self.orig_max_as_min = max_as_min
        else:
            self.orig_max_as_min = get_conf('max_as_min')
        self.update_spec()

        super().__init__(**kwargs)

    @default('cur_max_as_min')
    def _max_as_min_default(self):
        return self.orig_max_as_min

    @observe('cur_max_as_min')
    def update_max_as_min(self, change):
        self.update_spec()

    def update_spec(self):
        from desdeo_vis.plot.parallel import vega3_parplot_spec
        self.spec = vega3_parplot_spec(
            self.results, self.problem,
            max_as_min=self.cur_max_as_min,
            dim_tooltips=True, dim_symbols=True)


class NimbusPrefWidget(ParplotWidget):
    """
    A NIMBUS preference selection widget. This allows for graphical selection
    of preferences in the form NIMBUS requires.
    """

    _model_name = Unicode('NimbusPrefModel').tag(sync=True)
    _view_name = Unicode('NimbusPrefView').tag(sync=True)

    prefs = List(Dict()).tag(sync=True)
    prob = Unicode(allow_none=True).tag(sync=True)

    def __init__(self, results, problem, max_as_min=None, **kwargs):
        super().__init__(results, problem, max_as_min=max_as_min, **kwargs)

    def nimbus_clf(self, meth: NIMBUS) -> NIMBUSClassification:
        """
        Get the NIMBUS preference currently selected with the widget as a
        NIMBUSClassification. Raises a InvalidNimbusPreferencesException if an
        invalid preference is chosen.
        """
        if self.prob is not None:
            raise InvalidNimbusPreferencesException(
                "Preference is invalid: " + self.prob)
        return NIMBUSClassification(
            meth, [
                (
                    pref['kind'],
                    float(pref['val']) if 'val' in pref else 0.0
                )
                for pref in self.prefs
            ]
        )



#!/usr/bin/env python
# coding: utf-8

version_info = (0, 2, 0)
__version__ = ".".join(map(str, version_info))

# The version of the attribute spec that this package
# implements. This is the value used in
# _model_module_version/_view_module_version.
#
# Update this value when attributes are added/removed from
# your models, or serialized format changes.
EXTENSION_SPEC_VERSION = '1.0.0'



import sys
from os import mkdir
from os.path import isdir, join as pjoin
from shutil import copyfileobj

from pkg_resources import resource_stream, resource_listdir, resource_isdir


def main():
    output = False
    dest_dir = '.'

    for arg in sys.argv[1:]:
        if arg == '-o' or arg == '--output':
            output = True
        else:
            dest_dir = arg

    if not isdir(dest_dir):
        mkdir(dest_dir)

    desdeo_notebooks = 'desdeo_notebooks'
    if output:
        nb_dir = "output/"
    else:
        nb_dir = ""

    if output:
        print("Now run:")

    for nbf in resource_listdir(desdeo_notebooks, nb_dir):
        if not nbf.endswith('.ipynb'):
            continue
        path = nb_dir + nbf
        if resource_isdir(desdeo_notebooks, path):
            continue
        out_path = pjoin(dest_dir, nbf)
        copyfileobj(
            resource_stream(desdeo_notebooks, path),
            open(out_path, 'wb'))
        if output:
            print("jupyter trust {}".format(out_path))


if __name__ == '__main__':
    main()



"""
Globally configure default settings for plotting and widgets.
"""

_settings = {
    'max_as_min': True
}


def conf(**kwargs):
    """
    Set configuration values

    Parameters
    ----------
    max_as_min
        Whether to reformulate maximized functions as minimized functions.
    """
    for k, v in kwargs.items():
        _settings[k] = v


def get_conf(setting: str):
    """
    Get a configuration value

    Parameters
    ----------
    setting
        The configuration key to get.
    """
    return _settings[setting]



#!/usr/bin/env python
# coding: utf-8

from __future__ import print_function
from glob import glob
from os.path import join as pjoin


from setupbase import (
    create_cmdclass, install_npm, ensure_targets,
    find_packages, combine_commands, ensure_python,
    get_version, HERE
)

from setuptools import setup


# The name of the project
name = 'desdeo_vis'

# Ensure a valid python version
ensure_python('>=3.3')

# Get our version
version = get_version(pjoin(name, '_version.py'))

nb_path = pjoin(HERE, name, 'nbextension', 'static')
lab_path = pjoin(HERE, name, 'labextension')

# Representative files that should exist after a successful build
jstargets = [
    pjoin(nb_path, 'index.js'),
    pjoin(HERE, 'lib', 'plugin.js'),
]

package_data_spec = {
    name: [
        'nbextension/static/*.*js*',
        'labextension/*.tgz'
    ]
}

data_files_spec = [
    ('share/jupyter/nbextensions/desdeo_vis',
        nb_path, '*.js*'),
    ('share/jupyter/lab/extensions', lab_path, '*.tgz'),
    ('etc/jupyter/nbconfig/notebook.d', HERE, 'desdeo_vis.json')
]


cmdclass = create_cmdclass(
    'jsdeps',
    package_data_spec=package_data_spec,
    data_files_spec=data_files_spec)

cmdclass['jsdeps'] = combine_commands(
    install_npm(HERE, build_cmd='build:all'),
    #ensure_targets(jstargets),
)


setup_args = dict(
    name='desdeo-vis',
    description=(
        'Visualisations and preference selection widgets for the '
        + 'DESDEO interactive multiobjective optimization library'),
    version=version,
    scripts=glob(pjoin('scripts', '*')),
    cmdclass=cmdclass,
    packages=find_packages(),
    author='Frankie Robertson',
    author_email='frankie@robertson.name',
    url='https://github.com/industrial-optimization-group/desdeo-vis',
    license='MPL 2.0',
    platforms="Linux, Mac OS X, Windows",
    keywords=['Jupyter', 'Widgets', 'IPython'],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Framework :: Jupyter',
    ],
    include_package_data=True,
    install_requires=[
        'ipywidgets>=7.0.0',
        'widgetsnbextension',
        'desdeo>=0.1.3',
        'matplotlib',
        'altair',
        'vega',
    ],
    extras_require={
        'test': [
            'pytest',
            'pytest-cov',
            'nbval',
        ],
        'docs': [
            'sphinx',
            'recommonmark',
            'sphinx_rtd_theme',
            'nbsphinx>=0.2.13',
            'jupyter_sphinx',
            'nbsphinx-link',
            'pytest_check_links',
            'pypandoc',
            "sphinx_autodoc_typehints",
        ],
        "dev": ["black==18.4a4", "twine", "flake8", "isort"],
    },
    entry_points={},
    package_data={
        'desdeo_notebooks': [
            '*.ipynb',
            'output/*.ipynb',
        ]
    }
)

if __name__ == '__main__':
    setup(**setup_args)
